/**
 * @file kernel.hpp
 *
 * @brief Functions for querying information and making settings
 * regarding CUDA kernels (`__global__` functions).
 *
 * @note This file does _not_ define any kernels  itself.
 */
#pragma once
#ifndef CUDA_API_WRAPPERS_KERNEL_HPP_
#define CUDA_API_WRAPPERS_KERNEL_HPP_

#include <cuda/api/current_device.hpp>
#include <cuda/api/device_properties.hpp>
#include <cuda/api/error.hpp>
#include <cuda/common/types.hpp>

#include <cuda_runtime_api.h>

namespace cuda {

///@cond
class device_t;
class stream_t;
class kernel_t;
///@endcond

namespace kernel {

/**
 * Obtain a proxy object for a CUDA kernel
 *
 * @note This is a named constructor idiom, existing of direct access to the ctor
 * of the same signature, to emphasize that a new kernel is _not_ somehow created.
 *
 * @param device_id Device on which the texture is located
 * @param ptr pointer to the global device address of the kernel's executable code
 * @return a wrapper object associated with the specified kernel
 */
inline kernel_t wrap(device::id_t device_id, const void* ptr);

/**
 * @brief a wrapper around `cudaFuncAttributes`, offering
 * a few convenience member functions.
 */
struct attributes_t : cudaFuncAttributes {

	cuda::device::compute_capability_t ptx_version() const noexcept {
		return device::compute_capability_t::from_combined_number(ptxVersion);
	}

	cuda::device::compute_capability_t binary_compilation_target_architecture() const noexcept {
		return device::compute_capability_t::from_combined_number(binaryVersion);
	}
};

} // namespace kernel

/**
 * A non-owning wrapper class for CUDA `__global__` functions
 *
 * @note The association of a `kernel_t` with an individual device is somewhat tenuous.
 * That is, the same function pointer could be used with any other device (provided the kernel
 * was compiled appropriately). However, many/most of the features, attributes and settings
 * are device-specific.
 */
class kernel_t {
public: // getters
	const void* ptr() const noexcept { return ptr_; }
	device_t device() const noexcept;

protected:
	device::id_t device_id() const noexcept { return device_id_; }

public: // type_conversions
	operator const void*() noexcept { return ptr_; }

public: // non-mutators

	inline kernel::attributes_t attributes() const;

/*
	// The following are commented out because there are no CUDA API calls for them!
	// You may uncomment them if you'd rather get an exception...

	multiprocessor_cache_preference_t                cache_preference() const;
	multiprocessor_shared_memory_bank_size_option_t  shared_memory_bank_size() const;
*/

	/**
	 * @brief Calculates the number of grid blocks which may be "active" on a given GPU
	 * multiprocessor simultaneously (i.e. with warps from any of these block
	 * being schedulable concurrently)
	 *
	 * @param num_threads_per_block
	 * @param dynamic_shared_memory_per_block
	 * @param disable_caching_override On some GPUs, the choice of whether to
	 * cache memory reads affects occupancy. But what if this caching results in 0
	 * potential occupancy for a kernel? There are two options, controlled by this flag.
	 * When it is set to false - the calculator will assume caching is off for the
	 * purposes of its work; when set to true, it will return 0 for such device functions.
	 * See also the "Unified L1/Texture Cache" section of the
	 * <a href="http://docs.nvidia.com/cuda/maxwell-tuning-guide/index.html">Maxwell
	 * tuning guide</a>.
	 */
	grid::dimension_t maximum_active_blocks_per_multiprocessor(
		grid::block_dimension_t   num_threads_per_block,
		memory::shared::size_t    dynamic_shared_memory_per_block,
		bool                      disable_caching_override = false);

public: // mutators

	void set_attribute(kernel::attribute_t attribute, kernel::attribute_value_t value);

	/**
	 * @brief Change the hardware resource carve-out between L1 cache and shared memory
	 * for launches of the kernel to allow for at least the specified amount of
	 * shared memory.
	 *
	 * On several nVIDIA GPU micro-architectures, the L1 cache and the shared memory in each
	 * symmetric multiprocessor (=physical core) use the same hardware resources. The
	 * carve-out between the two uses has a device-wide value (which can be changed), but can
	 * also be set on the individual device-function level, by specifying the amount of shared
	 * memory the kernel may require.
	 */
	void opt_in_to_extra_dynamic_memory(cuda::memory::shared::size_t amount_required_by_kernel);

	/**
	 *
	 * @param dynamic_shared_memory_size The amount of dynamic shared memory each grid block will
	 * need.
	 * @param block_size_limit do not return a block size above this value; the default, 0,
	 * means no limit on the returned block size.
	 * @param disable_caching_override On platforms where global caching affects occupancy,
	 * and when enabling caching would result in zero occupancy, the occupancy calculator will
	 * calculate the occupancy as if caching is disabled. Setting this to true makes the
	 * occupancy calculator return 0 in such cases. More information can be found about this
	 * feature in the "Unified L1/Texture Cache" section of the
	 * <a href="https://docs.nvidia.com/cuda/maxwell-tuning-guide/index.html">Maxwell tuning guide</a>.
	 *
	 * @return A pair, with the second element being the maximum achievable block size
	 * (1-dimensional), and the first element being the minimum number of such blocks necessary
	 * for keeping the GPU "busy" (again, in a 1-dimensional grid).
	 */
	grid::complete_dimensions_t min_grid_params_for_max_occupancy(
		memory::shared::size_t   dynamic_shared_memory_size = no_dynamic_shared_memory,
		grid::block_dimension_t  block_size_limit = 0,
		bool                     disable_caching_override = false) const;

	template <typename UnaryFunction>
	grid::complete_dimensions_t min_grid_params_for_max_occupancy(
		UnaryFunction            block_size_to_dynamic_shared_mem_size,
		grid::block_dimension_t  block_size_limit = 0,
		bool                     disable_caching_override = false) const;

	/**
	 * @brief Indicate the desired carve-out between shared memory and L1 cache when launching
	 * this kernel - with fine granularity.
	 *
	 * On several nVIDIA GPU micro-architectures, the L1 cache and the shared memory in each
	 * symmetric multiprocessor (=physical core) use the same hardware resources. The
	 * carve-out between the two uses has a device-wide value (which can be changed), but the
	 * driver can set another value for a specific function. This function doesn't make a demand
	 * from the CUDA runtime (as in @p opt_in_to_extra_dynamic_memory), but rather indicates
	 * what is the fraction of L1 to shared memory it would like the kernel scheduler to carve
	 * out.
	 *
	 * @param shared_mem_percentage The percentage - from 0 to 100 - of the combined L1/shared
	 * memory space the user wishes to assign to shared memory.
	 *
	 * @note similar to @ref set_cache_preference() - but with finer granularity.
	 */
	void set_preferred_shared_mem_fraction(unsigned shared_mem_percentage);

	/**
	 * @brief Indicate the desired carve-out between shared memory and L1 cache when launching
	 * this kernel - with coarse granularity.
	 *
	 * On several nVIDIA GPU micro-architectures, the L1 cache and the shared memory in each
	 * symmetric multiprocessor (=physical core) use the same hardware resources. The
	 * carve-out between the two uses has a device-wide value (which can be changed), but the
	 * driver can set another value for a specific function. This function doesn't make a demand
	 * from the CUDA runtime (as in @p opt_in_to_extra_dynamic_memory), but rather indicates
	 * what is the fraction of L1 to shared memory it would like the kernel scheduler to carve
	 * out.
	 *
	 * @param preference one of: as much shared memory as possible, as much
	 * L1 as possible, or no preference (i.e. using the device default).
	 *
	 * @note similar to @ref set_preferred_shared_mem_fraction() - but with coarser granularity.
	 */
	void set_cache_preference(multiprocessor_cache_preference_t preference);

	/**
	 * @brief Sets a device function's preference of shared memory bank size preference
	 * (for the current device probably)
	 *
	 * @param config bank size setting to make
	 */
	void set_shared_memory_bank_size(multiprocessor_shared_memory_bank_size_option_t config);


protected: // ctors & dtor
	kernel_t(device::id_t device_id, const void* f)
	: device_id_(device_id), ptr_(f)
	{
		// TODO: Consider checking whether this actually is a device function
		// TODO: Consider performing a check for nullptr
	}

public: // ctors & dtor
	~kernel_t() = default;

	friend kernel_t kernel::wrap(device::id_t, const void* ptr);

protected: // data members
	const device::id_t device_id_;
	const void* const ptr_;
};

namespace kernel {

inline kernel_t wrap(device::id_t device_id, const void* function_ptr)
{
	return { device_id, reinterpret_cast<const void*>(function_ptr) };
}

template<typename KernelFunctionPtr>
kernel_t wrap(const device_t &device, KernelFunctionPtr function_ptr);

} // namespace kernel

} // namespace cuda

#endif // CUDA_API_WRAPPERS_KERNEL_HPP_
